#include <io.h>
#include <stdio.h>
#include <direct.h>
#include <windows.h>
#include <sys/stat.h>
#include <process.h>
#include <commctrl.h>
#include "mpc\in_mpc.h"
#include "mpc\requant.h"
#include "mpc\huffsv46.h"
#include "mpc\huffsv7.h"
#include "mpc\bitstream.h"
#include "mpc\mpc_dec.h"
#include "mpc\minimax.h"
#include "resource.h"
#include <tchar.h>

#ifndef _MAX_PATH
#define _MAX_PATH   260
#endif

#ifdef _UNICODE
# define OBJECT_NAME         _T("MPCscan U")
#else
# define OBJECT_NAME         _T("MPCscan A")
#endif

LPCTSTR TXT_SEP_LINE    = _T("----------\r\n");
LPCTSTR TXT_BT_SCAN     = _T("Scan");
LPCTSTR TXT_BT_STOP     = _T("Stop");
LPCTSTR TXT_DRAGHERE    = _T("Drag and drop files/directories you want to scan...");

HWND Dialog;
HANDLE thread_handle = INVALID_HANDLE_VALUE;

TCHAR **file_list;
UINT file_count;
UINT errors;
UINT id3v2tags;
unsigned __int64 totalsize;
unsigned __int64 bytes;
DWORD Cancel;
TCHAR *buffer;
TCHAR *ppos;

UINT ParamMessage;

// ------------------------------------

void FreeFileList ()
{
    for ( UINT i = 0; i < file_count; i++ ) {
        free ( file_list[i] );
    }
    free ( file_list );
    file_list  = NULL;
    file_count = 0;
}

UINT get_size ( LPCTSTR fn )
{
    HANDLE handle = CreateFile ( fn, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );
    if ( handle == INVALID_HANDLE_VALUE ) return 0;
    DWORD size = GetFileSize ( handle, NULL );
    CloseHandle ( handle );
    return size;
}

int isdir ( LPCTSTR path )
{
    DWORD attr = GetFileAttributes ( path );
    if ( attr == -1 ) return 0;
    return attr & FILE_ATTRIBUTE_DIRECTORY;
}

int AddFile ( LPCTSTR fname )
{
    LPCTSTR extensions[] = {
        _T(".mpc"), _T(".mp+"), _T(".mpp")
    };

    UINT i;
    BOOL is_mpc = FALSE;
    LPCTSTR p = _tcsrchr ( fname, _T('.') );
    if ( p == NULL ) return 0;

    for ( i = 0; i < sizeof(extensions)/sizeof(*extensions); i++ ) {
        if ( _tcsicmp (p, extensions[i]) == 0 ) {
            is_mpc = TRUE;
            break;
        }
    }
    if ( !is_mpc ) return 0;

    for ( i = 0; i < file_count; i++ ) {
        if ( _tcsicmp (fname, file_list[i]) == 0 ) {
            return 0;
        }
    }

    if ( (file_list = (TCHAR **)realloc ( file_list, sizeof (*file_list) * (file_count+1) )) == NULL ) {
        file_count = 0;
        return -1;
    }

    if ( (file_list[file_count] = (TCHAR *)malloc ((_tcslen (fname)+1) * sizeof (TCHAR))) == NULL ) {
        FreeFileList ();
        return -1;
    }

    _tcscpy ( file_list[file_count], fname );
    totalsize += get_size ( fname );
    file_count++;

    return 0;
}

void MakeFileName ( LPTSTR buffer, LPCTSTR path, LPCTSTR name )
{
    TCHAR *p = buffer;
    p += wsprintf ( p, _T("%s"), path );
    if ( p > buffer && *(p-1) != '\\' ) p += wsprintf ( p, _T("\\") );
    wsprintf ( p, _T("%s"), name );
}

int AddDir ( LPCTSTR path )
{
    TCHAR           temp[_MAX_PATH*2];
    WIN32_FIND_DATA FindFileData;
    HANDLE          hFind;

    wsprintf ( temp, _T("%s\\*"), path );
    hFind = FindFirstFile ( temp, &FindFileData );

    if ( hFind != INVALID_HANDLE_VALUE ) {
        if ( FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
            if ( _tcsicmp (FindFileData.cFileName, _T(".")) != 0 && _tcsicmp (FindFileData.cFileName, _T("..")) != 0 ) {
                MakeFileName ( temp, path, FindFileData.cFileName );
                if ( AddDir (temp) < 0 ) {
                    FindClose ( hFind );
                    return -1;
                }
            }
        } else {
            MakeFileName ( temp, path, FindFileData.cFileName );
            if ( AddFile (temp) < 0 ) {
                FindClose ( hFind );
                return -1;
            }
        }

        while ( FindNextFile (hFind, &FindFileData) ) {
            if ( FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
                if ( _tcsicmp (FindFileData.cFileName, _T(".")) != 0 && _tcsicmp (FindFileData.cFileName, _T("..")) != 0 ) {
                    MakeFileName ( temp, path, FindFileData.cFileName );
                    if ( AddDir (temp) < 0 ) {
                        FindClose ( hFind );
                        return -1;
                    }
                }
            } else {
                MakeFileName ( temp, path, FindFileData.cFileName );
                if ( AddFile (temp) < 0 ) {
                    FindClose ( hFind );
                    return -1;
                }
            }
        }

        FindClose ( hFind );
    }

    return 1;
}

void GetDroppedFiles ( HWND hWnd, HDROP hDrop )
{
    TCHAR fname[_MAX_PATH];
    UINT files = DragQueryFile ( hDrop, 0xFFFFFFFF, NULL, 0 );

    for ( UINT i = 0; i < files; i++ ) {
        DragQueryFile ( hDrop, i, fname, sizeof (fname) );

        if ( isdir (fname) ) {
            if ( AddDir (fname) < 0 ) {
                FreeFileList ();
                break;
            }
        } else {
            if ( AddFile (fname) < 0 ) {
                FreeFileList ();
                break;
            }
        }
	}

	DragFinish ( hDrop );
}

void AddCmdLineFile ( LPCTSTR buffer )
{
    TCHAR fname[_MAX_PATH];
    const TCHAR *s = buffer;
    TCHAR *d = fname;
    int quotes;

    if ( *s == '"' ) {
        quotes = 1;
        s++;
    } else {
        quotes = 0;
    }

    while ( *s ) {
        TCHAR ch = *s++;
        if ( quotes && ch == '"' ) break;
        if ( ch == ' ' && !quotes ) break;
        if ( d - fname < _MAX_PATH ) *d++ = ch;
    }
    *d = '\0';

    if ( isdir (fname) ) {
        if ( AddDir (fname) < 0 ) {
            FreeFileList ();
        }
    } else {
        if ( AddFile (fname) < 0 ) {
            FreeFileList ();
        }
    }
}

void GetCmdLineFiles ( LPTSTR buffer )
{
    TCHAR *argv;

    if ( buffer ) {
        argv = buffer;
    } else {
        argv = GetCommandLine ();
    }

    if ( argv == NULL ) return;
    int quotes = 0;
    int nextarg = 1;
    int argc = 0;

    while ( *argv ) {
        if ( nextarg ) {
            if ( buffer || (argc > 0) ) {
                AddCmdLineFile ( argv );
            }
            nextarg = 0;
            argc++;
        }

        if ( *argv == '"' ) quotes = 1 - quotes;
        if ( *argv == ' ' && !quotes ) nextarg = 1;
        argv++;
    }
}

int print_time ( LPTSTR str, unsigned int th, unsigned int tm, unsigned int ts )
{
    if ( th > 0 )
        return wsprintf ( str, _T("%02u:%02u:%02u"), th, tm, ts ); 
    else
        return wsprintf ( str, _T("%02u:%02u"), tm, ts ); 
}

MPC_decoder *theDec = NULL;
Reader* reader = NULL;
StreamInfo *Info = NULL;
UINT curfile = 0;

int OpenMPC ( UINT fnum )
{
    if ( fnum >= file_count ) return 1;

    HANDLE handle = CreateFile ( file_list[fnum], GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );

    if ( handle == INVALID_HANDLE_VALUE ) {
        ppos += wsprintf ( ppos, _T("%s: file open failed.\r\n"), file_list[fnum] );
        SetDlgItemText ( Dialog, IDC_EDIT1, buffer );
        errors++;
        return 1;
    }

    reader = new Reader_file ( handle );
    Info = new StreamInfo ();
    int ee;
    if ( (ee = Info->ReadStreamInfo (reader)) != 0 ) {
        ppos += wsprintf ( ppos, _T("%s: unrecognized format (%u).\r\n"), file_list[fnum], ee );
        SetDlgItemText ( Dialog, IDC_EDIT1, buffer );
        if ( reader ) { delete reader; reader = NULL; }
        if ( Info ) { delete Info; Info = NULL; }
        errors++;
        return 1;
    }

    theDec = new MPC_decoder ( reader );
    theDec->RESET_Globals   ();
    theDec->RESET_Synthesis ();
    theDec->SetStreamInfo ( Info );
    theDec->FileInit ();

    return 0;
}

void CloseMPC ()
{
    if ( Info   ) { delete Info;   Info   = NULL; }
    if ( reader ) { delete reader; reader = NULL; }
    if ( theDec ) { delete theDec; theDec = NULL; }
}

DWORD WINAPI ScanThread ( void *param )
{
    unsigned int lastpos = -1;
    unsigned int pos;

    while ( *(int*)param == 0 ) {
        if ( theDec == NULL ) {
            if ( OpenMPC (curfile) == 0 ) {
                if ( Info->simple.HeaderPosition > 0 ) {
                    if ( id3v2tags == 0 && errors == 0 ) {
                        ppos += wsprintf ( ppos, TXT_SEP_LINE );
                    }
                    ppos += wsprintf ( ppos, _T("%s: invalid MPC, contains ID3v2 tag!\r\n"), file_list[curfile] );
                    SetDlgItemText ( Dialog, IDC_EDIT1, buffer );
                    id3v2tags++;
                }
            } else {
                if ( curfile < file_count ) curfile++;
            }
        }

        if ( theDec ) {
            unsigned int RING = theDec->Zaehler;
            int valid_samples = theDec->DECODE ( NULL );
            theDec->UpdateBuffer ( RING );

            if ( theDec->FrameWasValid == 0 ) {
                TCHAR time1[32];
                unsigned long ts = (unsigned long)(theDec->DecodedFrames * FRAMELEN / Info->simple.SampleFreq);
                unsigned long tm = (ts/60) % 60;
                unsigned long th = (ts/60/60);
                ts = ts % 60;
                print_time ( time1, th, tm, ts );

                if ( id3v2tags == 0 && errors == 0 ) {
                    ppos += wsprintf ( ppos, TXT_SEP_LINE );
                }
                errors++;
                ppos += wsprintf ( ppos, _T("%s: broken after frame %u/%u (%s)\r\n"), file_list[curfile], theDec->DecodedFrames, Info->simple.Frames, time1 );
                SetDlgItemText ( Dialog, IDC_EDIT1, buffer );

                bytes += Info->simple.TotalFileLength;
                CloseMPC ();
                if ( curfile < file_count ) curfile++;
                continue;
            }
            if ( valid_samples <= 0 ) {
                bytes += Info->simple.TotalFileLength;
                CloseMPC ();
                if ( curfile < file_count ) curfile++;
                continue;
            }

            if ( totalsize > 0 ) {
                unsigned __int64 t = (bytes + (theDec->BitsRead()/8)) * 100 / totalsize;
                pos = (unsigned int)t;

                if ( pos != lastpos ) {
                    lastpos = pos;
                    if ( pos > 100 ) pos = 100;
                    SendDlgItemMessage ( Dialog, IDC_PROGRESS, PBM_SETPOS, pos, 0 );
                }
            }
        } else {
            if ( file_count > 0 && curfile >= file_count ) {
                CloseMPC ();
                UINT _files  = file_count;
                UINT _errors = errors;
                UINT _id3v2tags = id3v2tags;
                FreeFileList ();
                SendDlgItemMessage ( Dialog, IDC_PROGRESS, PBM_SETPOS, 0, 0 );
                SetDlgItemText ( Dialog, IDOK, TXT_BT_SCAN );
                ppos += wsprintf ( ppos, TXT_SEP_LINE );
                ppos += wsprintf ( ppos, _T("Scanned %u file"), _files );
                if ( _files != 1 ) ppos += wsprintf ( ppos, _T("s") );
                if ( _errors == 0 ) {
                    ppos += wsprintf ( ppos, _T(", no errors found.\r\n") );
                } else {
                    ppos += wsprintf ( ppos, _T(", errors with %u.\r\n"), _errors );
                }
                errors = 0;
                id3v2tags = 0;
                totalsize = 0;
                bytes = 0;
                curfile = 0;
                SetDlgItemText ( Dialog, IDC_EDIT1, buffer );
                EnableWindow ( GetDlgItem (Dialog, IDOK), FALSE );
                Cancel = 1;
                DragAcceptFiles ( Dialog, TRUE );
            }
        }
    }

    return 0;
}

BOOL CALLBACK DialogProc ( HWND hWndDlg, UINT Message, WPARAM wParam, LPARAM lParam )
{
    HDROP hDrop;
    DWORD thread_id = 0;

    switch ( Message ) {
    case WM_INITDIALOG:
        DragAcceptFiles ( hWndDlg, TRUE );
        SendDlgItemMessage ( hWndDlg, IDC_PROGRESS, PBM_SETRANGE, 0, MAKELONG(0, 100) );
        SendDlgItemMessage ( hWndDlg, IDC_PROGRESS, PBM_SETPOS,   0, 0 );
        SendDlgItemMessage ( hWndDlg, IDC_PROGRESS, PBM_SETSTEP,  1, 0 );
        SetDlgItemText ( hWndDlg, IDOK, TXT_BT_SCAN );
        SetDlgItemText ( hWndDlg, IDC_EDIT1, TXT_DRAGHERE );
        UpdateWindow ( hWndDlg );
        EnableWindow ( GetDlgItem (hWndDlg, IDOK), FALSE );
        if ( file_count > 0 ) {
            TCHAR temp[256];
            TCHAR *p = temp;
            p += wsprintf ( p, _T("%u file"), file_count );
            if ( file_count != 1 ) p += wsprintf ( p, _T("s") );
            wsprintf ( p, _T(" waiting for scanning...") );
            SetDlgItemText ( hWndDlg, IDC_EDIT1, temp );
            EnableWindow ( GetDlgItem (hWndDlg, IDOK), TRUE );
            SetFocus ( GetDlgItem (hWndDlg, IDOK) );
        }
        break;

    case WM_CLOSE:
        DestroyWindow ( hWndDlg );
        PostQuitMessage ( 0 );
        return TRUE;

	case WM_DROPFILES:
        if ( Cancel ) {
		    hDrop = (HDROP)wParam;
            UINT c = file_count;
            GetDroppedFiles ( hWndDlg, hDrop );
            if ( file_count > c ) {
                TCHAR temp[256];
                TCHAR *p = temp;
                p += wsprintf ( p, _T("%u file"), file_count );
                if ( file_count != 1 ) p += wsprintf ( p, _T("s") );
                wsprintf ( p, _T(" waiting for scanning...") );
                SetDlgItemText ( hWndDlg, IDC_EDIT1, temp );
                if ( c == 0 ) {
                    EnableWindow ( GetDlgItem (hWndDlg, IDOK), TRUE );
                    SetFocus ( GetDlgItem (hWndDlg, IDOK) );
                }
            }
            return 0;
        }
        break;

    case WM_COPYDATA:
        if ( Cancel && wParam == 0 ) {
            COPYDATASTRUCT *cdata = (COPYDATASTRUCT *)lParam;
            if ( cdata->cbData > 0 ) {
                UINT c = file_count;
                GetCmdLineFiles ( (TCHAR *)cdata->lpData );
                if ( file_count > c ) {
                    TCHAR temp[256];
                    TCHAR *p = temp;
                    p += wsprintf ( p, _T("%u file"), file_count );
                    if ( file_count != 1 ) p += wsprintf ( p, _T("s") );
                    wsprintf ( p, _T(" waiting for scanning...") );
                    SetDlgItemText ( hWndDlg, IDC_EDIT1, temp );
                    if ( c == 0 ) {
                        EnableWindow ( GetDlgItem (hWndDlg, IDOK), TRUE );
                        SetFocus ( GetDlgItem (hWndDlg, IDOK) );
                    }
                }
            }
        }
        break;

    case WM_COMMAND:
        switch ( LOWORD (wParam) ) {
        case IDOK:
            if ( Cancel ) {
                DragAcceptFiles ( hWndDlg, FALSE );

                if ( thread_handle != INVALID_HANDLE_VALUE ) {
                    if ( WaitForSingleObject ( thread_handle, INFINITE ) == WAIT_TIMEOUT )
                        TerminateThread ( thread_handle, 0 );

                    CloseHandle ( thread_handle );
                    thread_handle = INVALID_HANDLE_VALUE;
                }

                if ( buffer ) free ( buffer );
                if ( (buffer = (TCHAR *)malloc ((256 + file_count * (_MAX_PATH + 128)) * sizeof (TCHAR))) == NULL ) {
                    DestroyWindow ( hWndDlg );
                    PostQuitMessage ( 0 );
                    return TRUE;
                }

                ppos = buffer;
                ppos += wsprintf ( ppos, _T("Scanning %u file"), file_count );
                if ( file_count != 1 ) ppos += wsprintf ( ppos, _T("s") );
                ppos += wsprintf ( ppos, _T("...\r\n") );
                SetDlgItemText ( hWndDlg, IDOK, TXT_BT_STOP );
                SetDlgItemText ( hWndDlg, IDC_EDIT1, buffer );
                UpdateWindow ( hWndDlg );
                Cancel = 0;
                thread_handle = (HANDLE)CreateThread ( NULL, 0, (LPTHREAD_START_ROUTINE)ScanThread, (void*)&Cancel, 0, &thread_id );
                SetThreadPriority ( thread_handle, THREAD_PRIORITY_IDLE );
            } else {
                Cancel = 1;
                SetDlgItemText ( hWndDlg, IDOK, TXT_BT_SCAN );
                FreeFileList ();
                errors = 0;
                id3v2tags = 0;
                totalsize = 0;
                bytes = 0;
                curfile = 0;
                SetDlgItemText ( hWndDlg, IDC_EDIT1, TXT_DRAGHERE );
                SendDlgItemMessage ( hWndDlg, IDC_PROGRESS, PBM_SETPOS, 0, 0 );
                EnableWindow ( GetDlgItem (hWndDlg, IDOK), FALSE );
                DragAcceptFiles ( hWndDlg, TRUE );
                if ( thread_handle != INVALID_HANDLE_VALUE ) {
                    if ( WaitForSingleObject ( thread_handle, INFINITE ) == WAIT_TIMEOUT )
                        TerminateThread ( thread_handle, 0 );

                    CloseHandle ( thread_handle );
                    thread_handle = INVALID_HANDLE_VALUE;
                }
                CloseMPC ();
            }
            break;
        }
        break;

    default:
        return FALSE;
    }

    return TRUE;
}

void setpriority ()
{
    HANDLE  pid = GetCurrentProcess ();
    HANDLE  tid = GetCurrentThread  ();

    SetPriorityClass ( pid, IDLE_PRIORITY_CLASS );
    SetThreadPriority ( tid, THREAD_PRIORITY_IDLE );
}

int WINAPI WinMain ( HINSTANCE Instance, HINSTANCE, PSTR szCmdLine, int ShowCmd )
{
    file_list  = NULL;
    file_count = 0;
    errors = 0;
    id3v2tags = 0;
    totalsize = 0;
    bytes = 0;
    Cancel = 1;
    buffer = NULL;

    HANDLE hMutex;
    hMutex = OpenMutex ( MUTEX_ALL_ACCESS, FALSE, OBJECT_NAME );
    if ( hMutex ) {
        if ( szCmdLine && szCmdLine[0] ) {
            HWND hWnd = FindWindow ( NULL, _T("MPCscan") );
            if ( hWnd == NULL ) return 0;

            LPTSTR argv = GetCommandLine ();
            if ( argv == NULL ) return 0;

            TCHAR *p = argv;
            int quotes;

            if ( *p == '"' ) {
                p++;
                quotes = 1;
            } else {
                quotes = 0;
            }
            while ( *p ) {
                TCHAR ch = *p++;
                if ( quotes && ch == '"' ) break;
                if ( ch == ' ' && !quotes ) break;
            }
            if ( *p == ' ' ) p++;
            if ( *p == '\0' ) p = argv;

            COPYDATASTRUCT cdata;
            cdata.dwData = 1;
            cdata.cbData = (_tcslen (p) + 1) * sizeof(TCHAR);
            cdata.lpData = (unsigned long *)p;
            SendMessage ( hWnd, WM_COPYDATA, 0, (LPARAM)&cdata );
        }
        return 0;
    }
    hMutex = CreateMutex ( NULL, TRUE, OBJECT_NAME );

    setpriority ();
    GetCmdLineFiles ( NULL );

    MSG Msg;
    HWND Dlg = CreateDialog ( Instance, MAKEINTRESOURCE(IDD_DIALOG), 0, DialogProc );
    Dialog = Dlg;
    RECT Rect;
    if ( GetWindowRect (Dlg, &Rect) ) {
	    SetWindowPos ( Dlg, HWND_TOPMOST, Rect.left, Rect.top, Rect.right, Rect.bottom, 0 );
    }
    ShowWindow ( Dlg, ShowCmd );

    UpdateWindow (Dlg );

    while ( GetMessage (&Msg, 0, 0, 0) ) {
        if ( !IsDialogMessage (Dlg, &Msg) ) {
            TranslateMessage ( &Msg );
            DispatchMessage ( &Msg );
        }
    }

    ReleaseMutex ( hMutex );

    return Msg.wParam;
}
